// 6.2.6 - matched-filter core calculation for the CUDA kernel

const double c = 299792458.0; // speed of light

// the core of the Matched filter algorithm 
__global__ void mf_core(double2 *out, const double *dR, 
           const double *freq, const double2 *phdata, const int numel)
{
    // calculate a global thread index within a matrix dR
    int idx = blockIdx.x * blockDim.x + threadIdx.x;
    if (idx >= numel)  // early bail-out for unused threads 
        return;

    double f = freq[blockIdx.y]; 
    double2 ph = phdata[blockIdx.y], r;
    double t = 4.0 * M_PI * f * dR[idx] / c, re, im;
    
    sincos(t, &im, &re);
    r.x = ph.x*re + ph.y*im;  // multiply two complex numbers
    r.y = ph.x*im - ph.y*re;
    idx = idx + blockIdx.y * numel;
    out[idx] = r;    
}
